#include <cstdio>
#include <FL/Fl.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Input.H>
#include <FL/Fl_Double_Window.H>
#include <FL/Fl_Menu_Window.H>


/* @FIXME some controls are broken with this enabled */
#define USE_GRAB 1


const static char * E[] = {
	"FL_NO_EVENT",
	"FL_PUSH",
	"FL_RELEASE",
	"FL_ENTER",
	"FL_LEAVE",
	"FL_DRAG",
	"FL_FOCUS",
	"FL_UNFOCUS",
	"FL_KEYDOWN",
	"FL_KEYUP",
	"FL_CLOSE",
	"FL_MOVE",
	"FL_SHORTCUT",
	"FL_DEACTIVATE",
	"FL_ACTIVATE",
	"FL_HIDE",
	"FL_SHOW",
	"FL_PASTE",
	"FL_SELECTIONCLEAR",
	"FL_MOUSEWHEEL",
	"FL_DND_ENTER",
	"FL_DND_DRAG",
	"FL_DND_LEAVE",
	"FL_DND_RELEASE",
	"FL_SCREEN_CONFIGURATION_CHANGED",
	"FL_FULLSCREEN",
	"FL_ZOOM_GESTURE"
};


class Input : public Fl_Input {
public:
	explicit Input(const int X, const int Y, const int W, const int H, const char * l = NULL):
		Fl_Input(X, Y, W, H, l)
	{}
	
	virtual ~Input() {}
	
	virtual int handle(int e) {
		printf(
			"Input(%s): %ix%i %ix%i %s == %s  '%.*s'\n", E[e],
			Fl::event_x(), Fl::event_y(),
			Fl::event_x_root(), Fl::event_y_root(),
			Fl::event_inside(x(), y(), w(), h()) ? "true" : "false",
			Fl::event_inside(this) ? "true" : "false",
			Fl::event_length(), Fl::event_text()
		);
		return Fl_Input::handle(e);
	}
};


class PopupWindow : public Fl_Menu_Window {
private:
	Input * input;
	Fl_Button * ok;
	bool done;
public:
	explicit PopupWindow():
		Fl_Menu_Window(180, 130)
	{
		clear_border();
		set_modal();
		box(FL_UP_BOX);
		/* @FIXME navigation does not work in group of group with grabbing */
		//Fl_Group * group = new Fl_Group(0, 0, w(), h());
		//group->box(FL_UP_BOX);
			new Fl_Box(10, 10, w() - 20, 20, "Input..");
			input = new Input(10, 40, w() - 20, 20);
			input->readonly(0);
			input->insert("write here");
			new Fl_Button(10, 70, w() - 20, 20, "nothing");
			ok = new Fl_Button(10, 100, w() - 20, 20, "OK");
			ok->callback(onOk, this);
		//group->end();
		end();
	}
	
	virtual ~PopupWindow() {}
	
	virtual int handle(int e) {
		int result = 0;
		/* @FIXME workaround for keyboard event handling if grabbing is on (see Fl_Group::handle()) */
		if (e == FL_KEYBOARD && Fl::focus() && Fl::focus()->window() == this) {
			result = Fl::focus()->handle(e);
			if ( ! result ) result = Fl_Menu_Window::handle(e);
		} else {
			result = Fl_Menu_Window::handle(e);
		}
		//printf(
		//	"Popup(%s): %ix%i %ix%i %s == %s\n", E[e],
		//	Fl::event_x(), Fl::event_y(),
		//	Fl::event_x_root(), Fl::event_y_root(),
		//	Fl::event_inside(0, 0, w(), h()) ? "true" : "false",
		//	Fl::event_inside(child(0)) ? "true" : "false"
		//);
		switch (e) {
		case FL_PUSH:
		case FL_DRAG:
		case FL_RELEASE:
			if ( ! Fl::event_inside(0, 0, w(), h()) ) {
				hide();
				result = 1;
			}
			break;
		case FL_KEYBOARD:
			switch (Fl::event_key()) {
			case FL_Escape:
			case FL_KP_Enter:
			case FL_Enter:
				hide();
				result = 1;
			default:
				break;
			}
			break;
		case FL_HIDE:
			done = true;
			result = 1;
			break;
		default:
			break;
		}
		return result;
	}
	
	/* implemented like in fl_show_colormap.cxx */
	void show(const int X, const int Y) {
#ifdef USE_GRAB
		Fl::grab(*this);
#endif /* USE_GRAB */
		position(X, Y);
		Fl_Menu_Window::show();
		input->take_focus();
#ifdef USE_GRAB
		done = false;
		while ( ! done ) Fl::wait();
		Fl::grab(0);
#endif /* USE_GRAB */
	}
	
	static void onOk(Fl_Widget * w, void * user) {
		if (w == NULL || user == NULL) return;
		PopupWindow * self = reinterpret_cast<PopupWindow *>(user);
		self->hide();
	}
};


class MainWindow : public Fl_Double_Window {
private:
	Fl_Button * button;
	PopupWindow * popup;
public:
	explicit MainWindow(const int W, const int H):
		Fl_Double_Window(W, H, "main")
	{
		button = new Fl_Button(W * 1 / 4, H * 1 / 4, W / 2, H / 2, "show popup");
		button->callback(onButtonClick, this);
		end();
		popup = new PopupWindow();
	}
	
	virtual ~MainWindow() {
		delete popup;
	}
	
	virtual int handle(int e) {
		const int result = Fl_Double_Window::handle(e);
		if (e == FL_HIDE) popup->hide();
		return result;
	}
	
	void showPopup() const {
		popup->show(x() + button->x(), y() + button->y() + button->h());
	}
private:
	static void onButtonClick(Fl_Widget * w, void * user) {
		if (w == NULL || user == NULL) return;
		MainWindow * self = reinterpret_cast<MainWindow *>(user);
		self->showPopup();
	}
};


int main() {
	Fl::visual(FL_DOUBLE | FL_RGB);
	Fl::set_color(FL_BACKGROUND_COLOR, 212, 208, 200);
	MainWindow window(240, 120);
	window.show();
	return Fl::run();
}
